import java.util.*;
import java.io.*;
import java.io.Serializable;
import java.util.Vector;

import iBus.*;
import iBus.layers.*;
import iBus.exception.*;

import Ticker.*;

/* ***************************************************************************
 * Classe: Archiver
 *
 */
public class Archiver {
  static String appName = "Archiver";
  static Log log = new Log(appName);			// Objecto log
  static String channelVName = "Tick-Channel";
  static String channelVVer = "1.5";
  static String channelVSubject = "/P";
  static String channelVQoS = "Reliable";

  static String channelAName = channelVName;
  static String channelAVer = channelVVer;
  static String channelASubject = "/A";
  static String channelAQoS = "Reliable";

  static MembershipMan mm= new MembershipMan();		// Gestor de filiacao

  public static void main(String argv[]) throws Exception {
    Application.setAppName(appName);

    /* ***********************************************************************
     * Instancia a stack para o canal volatil (V).
     *
     * Os membros do canal volatil sao os Produtores (talkers),
     * os Consumidores (listeners) e os membros do Arquivo (listeners).
     */

    // Pilha de protocolos para difusao fiavel (reliable multicast)
    final iBus.Stack stackV = new iBus.Stack(channelVQoS);

    // Aumenta os valores do timeout e heartbeat para tornar a deteccao de
    // falhas menos agressiva (valores em milisegundos). Por outro lado,
    // assumimos que este canal nao perde mensagens.
    REACH reachV = (REACH)stackV.getProtocolObject("REACH");
    reachV.setTimeout(60000);
    reachV.setHbeat(10000);
    NAK nakV = (NAK)stackV.getProtocolObject("NAK");
    nakV.setHBInterval(60000);

    iBusURL urlV = null;				// Endereco do canal V
    try {
      urlV = iBusURLFactory.create(channelVName, channelVVer, channelVSubject);
    }
    catch (Exception e) {
      // Coloca no log uma mensagem e termina a aplicacao (exit code=1)
      log.panic("main: iBusURLFactory.create failed for channel P: " + e);
    }


    /* ***********************************************************************
     * Define o Aquivo como um consumidor do canal V.
     *
     * O Arquivo participa no canal V apenas como consumidor. Todas as
     * mensagens recebidas em V sao gravadas pelo Arquivo em memoria nao
     * volatil.
     */

    // Instancia o objecto que recebe as mensagens enviadas em V
    ObjectArchiver rcvV = new ObjectArchiver(log, mm, false);

    // Conecta-se ao canal V como assinante (provoca uma mudanca de vista em V)
    try {
      stackV.subscribe(urlV, rcvV);
    }
    catch (Exception e) {
      log.panic("main: stack.subscribe failed for channel P: " + e);
    }


    /* ***********************************************************************
     * Instancia a stack para o canal do aquivo (A).
     *
     * As mensagens recebidas em V sao mais tarde retransmitidas em A para os
     * consumidores do canal nao receberam essas mensagens.
     * O canal A serve para sincronizar os Consumidores mais atrasados com
     * os restantes membros do canal V.
     *
     * Os membros do canal do arquivo sao Consumidores (listeners),
     * os proprios membros do Arquivo (talkers/listners). A razao pelo qual
     * os membros do Arquivo tambem participam no canal A como consumidores
     * deve-se ao facto de o proprio Arquivo utilizar este canal para se
     * sincronizar com os restantes membros do arquivo no caso de falha.
     */

    final iBus.Stack stackA = new iBus.Stack(channelAQoS);

    // Aumenta os valores do timeout e heartbeat para tornar a deteccao de
    // falhas menos agressiva (valores em milisegundos). Tal como em V, tambem
    // assumimos que o canal A nao perde mensagens.
    REACH reachA = (REACH)stackA.getProtocolObject("REACH");
    reachA.setTimeout(60000);
    reachA.setHbeat(10000);
    NAK nakA = (NAK)stackA.getProtocolObject("NAK");
    nakA.setHBInterval(60000);

    iBusURL urlA = null;				// Endereco do canal A
    try {
      urlA = iBusURLFactory.create(channelAName, channelAVer, channelASubject);
    }
    catch (Exception e) {
      // Coloca no log uma mensagem e termina a aplicacao (exit code=1)
      log.panic("main: iBusURLFactory.create failed for channel A: " + e);
    }


    /* ***********************************************************************
     * Define o Aquivo como um consumidor do canal A.
     *
     * O Arquivo participa no canal A como consumidor para que, utilizando
     * o mesmo mecanismo dos Consumidores, se possa sincronizar com os
     * restantes membros do Arquivo no caso de perder alguma mensagem.
     */

    // Instancia o objecto que recebe as mensagens enviadas em A
    ObjectArchiver rcvA = new ObjectArchiver(log, mm, true);

    // Conecta-se ao canal A como assinante (provoca uma mudanca de vista em A)
    try {
      stackA.subscribe(urlA, rcvA);
    }
    catch (Exception e) {
      log.panic("main: stack.subscribe failed for channel A: " + e);
    }


    /* ***********************************************************************
     * Define o Arquivo como um produtor do canal A.
     *
     * As mensagens recebidas em P sao retransmitidas pelo Arquivo no canal A
     * para os Consumidores que falham.
     */

    // Conecta-se ao canal A como editor (provoca uma mudanca de vista em A)
    try {
      stackA.registerTalker(urlA);
    }
    catch (Exception e) {
      log.panic("main: stack.registerTalker failed for channel A: " + e);
    }


    /* ***********************************************************************
     * Regista o Arquivo como "monitor de filiacao" no canal V.
     *
     * Desta forma o Arquivo consegue detectar as entradas e saidas do canal
     * V de forma a poder gerir mais facilmente o processo de sincronizacao
     * dos Consumidores.
     */

    // Instancia o monitor de filiacao em V
    MembershipMon monV = new MembershipMon(mm,
		channelVSubject, channelASubject);

    // Regista o monitor de filiacao
    try {
      stackV.registerMonitor(urlV, monV);
    }
    catch (Exception e) {
      log.panic("main: stack.registerMonitor failed for channel P: " + e);
    }

    /* ***********************************************************************
     * Regista o Arquivo como "monitor de filiacao" no canal A.
     *
     * Desta forma o Arquivo consegue detectar as entradas e saidas do canal
     * A de forma a poder gerir mais facilmente o processo de sincronizacao
     * dos Consumidores.
     */

    // Instancia o monitor de filiacao em A
    MembershipMon monA = new MembershipMon(mm,
		channelVSubject, channelASubject);

    // Regista o monitor de filiacao
    try {
      stackA.registerMonitor(urlA, monA);
    }
    catch (Exception e) {
      log.panic("main: stack.registerMonitor failed for channel A: " + e);
    }


    /* ***********************************************************************
     * Inicio da aplicacao
     *
     */

    System.out.println(Application.getAppName() + ": inicio da aplicacao.\n");

    stackV.waitTillExit();			// Aguarda por mensagens em V
  }
}



/* ***************************************************************************
 * Classe: ObjectArchiver
 *
 */
class ObjectArchiver implements iBus.Receiver {
  Log log;		// Objecto log
  MembershipMan mm;	// Gestor de filiacao
  boolean archiver;	// Canal do arquivo (false=nao; true=sim)
  static int count = 0;

  // Constructor
  public ObjectArchiver(Log log_, MembershipMan mm_, boolean archiver_) {
    log = log_;
    mm = mm_;
    archiver = archiver_;
  }

  // O metodo dispatchPush e' invocado quando uma mensagem e' recebida
  public void dispatchPush(iBusURL channel, Posting posting) {
    if (!archiver) {
    // Se a mensagem foi recebida no canal dos produtores deve ser
    // tornada persistente (um ficheiro por cada mensagem recebida)
      for (int n = 0; n < posting.getLength(); n++) {
        String postingName = posting.getObject(n).getClass().getName();

        if (postingName.substring(0, 7).equals("Ticker.")) {
          Tick post = (Tick)posting.getObject(n);
          String sender = post.getSenderId();
          int order = post.getOrderNumber();

          if (SavePost(post, sender, order)) {
            mm.setMessageProducer(sender, new Integer(order));
          }
          else {
            //erro a gravar a mensagem
          }

          System.out.println((count++) + "\t" + new Date().toString());
        }
        else {
          String str = (String)posting.getObject(n);
          System.out.println("Mensagem desconhecida: " + str.toString());
        }
      }
    }

    System.out.flush();
  }

  // O metodo dispatchPull e' invocado quando uma mensagem e' enviada
  public Posting dispatchPull(iBusURL channel, Posting request) {
    System.err.println("ObjectArchiver.dispatchPull called");
    return(null);
  }

  // O metodo error e' invocado quando ocorre um erro
  public void error(iBusURL channel, String details) {
    log.panic("ObjectArchiver: internal error: " + details);
  }

  // O metodo SavePost grava as mensagens recebida em V
  private boolean SavePost(Object post, String sender, int order) {
    boolean status = false;

    try {
      ObjectOutputStream out = new ObjectOutputStream(
		new FileOutputStream("archiver-data/" + sender + "_" + order));
      out.writeObject(post);
      out.close();
      status = true;
    }
    catch (Exception e) {
      log.panic("ObjectArchiver.dispatchPush error: " + e);
    }

    return(status);
  }
}



/* ***************************************************************************
 * Classe: MembershipMon
 *
 * Esta classe e' utilizada para observar as alteracoes nas vistas do canal.
 */
class MembershipMon implements iBus.Membership {
  MembershipMan mm;
  String channelVSubject;
  String channelASubject;

  // Constructor
  public MembershipMon(MembershipMan mm_,
                String channelVSubject_, String channelASubject_) {
    channelVSubject = channelVSubject_;
    channelASubject = channelASubject_;
    mm = mm_;
  }

  public Serializable getState(iBusURL channel) {
    System.err.println("MembershipMon.getState called");
    return(null);
  }

  public void setState(iBusURL channel, Serializable state) {
    System.err.println("MembershipMon.setState called");
  }

  private boolean isChannelArchiver(String channel) {
    return(channelASubject.equals(channel.substring(channel.length() - channelASubject.length())));
  }

  public void viewChange(View newView) {
    String channel = newView.getChannel().getSubject();
    boolean channelA = isChannelArchiver(channel);	// true = canal A
    Vector members = new Vector();

    if (channelA) {		// Se ocorreu uma mudanca de vista no canal A
      for (int i = 0; i < newView.getMembers().length; i++) {
        ChannelMember m = newView.getMembers()[i];
        String address = m.getURL().getAddress();

        if (m.isTalker()) {			// Se e' um membro do arquivo
          // Atraves deste mecanismo podemos saber quais os membros do arquivo
        }
        else {				// Senao, e' um assinante persistente
          members.addElement(address);
        }
      }
    }
    else {					// Se vista mudou no canal V
      for (int i = 0; i < newView.getMembers().length; i++) {
        ChannelMember m = newView.getMembers()[i];
        String address = m.getURL().getAddress();

        if (m.isListener()) {				// Se e' um assinante
          members.addElement(address);
        }
      }
    }

    mm.updateConsumerList(channelA, members);
  }
}



/* ***************************************************************************
 * Classe: MembershipMan
 *
 * Esta classe e' utilizada para gerir as alteracoes nas vistas do canal.
 */
class MembershipMan {
  Vector membersP = new Vector();
  Vector membersA = new Vector();
  Vector messages = new Vector();
  Vector messagesView = new Vector();
  Vector consumersFailed = new Vector();
  Vector consumersFailedMessages = new Vector();

  // Faz a gestao da lista de editores do canal V
  public void setMessageProducer(String address, Integer order) {
    // A lista de editores contem o par (id_editor;nr_ordem_ultima_msg_recebida)
    if (membersP.size() > 0) {
      messagesView.addElement((address + "_" + order));
    }
  }

  public void updateConsumerList(boolean channelA, Vector members) {
    if (channelA) {
      membersA.addElement(members);

      checkMessagesToDeliver(members);
    }
    else {
      membersP.addElement(members);
      messages.addElement(messagesView);

      checkFailedMessages(members, messagesView);
    }
  }

  private void checkFailedMessages(Vector members, Vector messages) {
    messagesView = new Vector();

    for (int n = 0; n < members.size(); n++) {
      String address = (String)members.elementAt(n);

      if (!isMemberOnLastView(address)) {
      // Se o consumidor nao pertencia a vista anterior,
      // deve receber as mensagens dessa vista
        consumersFailed.addElement(address);
        consumersFailedMessages.addElement(messages);
      }

      // Mensagens incertas: todas as mensagens da vista corrente
      consumersFailed.addElement(address);
      consumersFailedMessages.addElement(messagesView);
    }
  }

  private void checkMessagesToDeliver(Vector members) {
    for (int n = 0; n < members.size(); n++) {
      String address = (String)members.elementAt(n);
 
      while (consumersFailed.contains(address)) {
        int index = consumersFailed.indexOf(address);

        Vector messages = (Vector)consumersFailedMessages.elementAt(index);

        if (messages.size() > 0) {
          System.err.println("O consumidor " + address +
			" deve receber as seguintes mensagens:");

          for (int m = 0; m < messages.size(); m++) {
            String message = (String)messages.elementAt(m);
            System.err.println("Mensagem " + message);
          }
        }

        consumersFailed.removeElementAt(index);
        consumersFailedMessages.removeElementAt(index);
      }
    }
  }

  private boolean isMemberOnLastView(String address) {
    boolean isMember = true;

    if (membersP.size() > 1) {
      Vector members = (Vector)membersP.elementAt(membersP.size() - 2);
      isMember = members.contains(address);
    }

    return(isMember);
  }
}

// fim: Archiver.java
