import java.util.*;

import iBus.*;
import iBus.layers.*;
import iBus.exception.*;

import Ticker.*;

/* ***************************************************************************
 * Classe: Consumer (persistente)
 *
 */
public class Consumer {
  static String appName = "Consumer";
  static Log log = new Log(appName);			// Objecto log
  static String channelVName = "Tick-Channel";
  static String channelVVer = "1.5";
  static String channelVSubject = "/P";
  static String channelVQoS = "Reliable";
  static String address;

  static String channelAName = channelVName;
  static String channelAVer = channelVVer;
  static String channelASubject = "/A";
  static String channelAQoS = "Reliable";

  public static void main(String argv[]) throws Exception {
    Application.setAppName(appName);

    /* ***********************************************************************
     * Instancia a stack para o canal volatil (V).
     *
     * Os membros do canal volatil sao os Produtores (talkers),
     * os Consumidores (listeners) e os membros do Arquivo (listeners).
     */

    // Pilha de protocolos para difusao fiavel (reliable multicast)
    final iBus.Stack stackV = new iBus.Stack(channelVQoS);

    // Aumenta os valores do timeout e heartbeat para tornar a deteccao de
    // falhas menos agressiva (valores em milisegundos). Por outro lado,
    // assumimos que este canal nao perde mensagens.
    REACH reachV = (REACH)stackV.getProtocolObject("REACH");
    reachV.setTimeout(60000);
    reachV.setHbeat(10000);
    NAK nakV = (NAK)stackV.getProtocolObject("NAK");
    nakV.setHBInterval(60000);

    iBusURL urlV = null;				// Endereco do canal V
    try {
      urlV = iBusURLFactory.create(channelVName, channelVVer, channelVSubject);
    }
    catch (Exception e) {
      // Coloca no log uma mensagem e termina a aplicacao (exit code=1)
      log.panic("main: iBusURLFactory.create failed for channel P: " + e);
    }


    // Define o endereco logico que permite identificar o produtor no canal
    address = urlV.getAddress();

    /* ***********************************************************************
     * Define a aplicacao como um consumidor do canal V.
     *
     * A aplicacao participa nos canais A e V como consumidor. As mensagens
     * nao recebidas em V sao depois solicitadas ao Arquivo e recebidas em A.
     */

    // Instancia o objecto que recebe as mensagens enviadas em V
    ObjectReceiver rcvV = new ObjectReceiver(log, address);

    // Conecta-se ao canal V como assinante (provoca uma mudanca de vista em V)
    try {
      stackV.subscribe(urlV, rcvV);
    }
    catch (Exception e) {
      log.panic("main: stack.subscribe failed for channel P: " + e);
    }


    /* ***********************************************************************
     * Instancia a stack para o canal do aquivo (A).
     *
     * Este canal e utilizado para receber as mensagens nao recebidas em V.
     * Desta forma o Consumidor pode ser sincronizado com os restantes
     * consumidores do canal V.
     */

    final iBus.Stack stackA = new iBus.Stack(channelAQoS);

    // Aumenta os valores do timeout e heartbeat para tornar a deteccao de
    // falhas menos agressiva (valores em milisegundos). Tal como em V, tambem
    // assumimos que o canal A nao perde mensagens.
    REACH reachA = (REACH)stackA.getProtocolObject("REACH");
    reachA.setTimeout(60000);
    reachA.setHbeat(10000);
    NAK nakA = (NAK)stackA.getProtocolObject("NAK");
    nakA.setHBInterval(60000);

    iBusURL urlA = null;				// Endereco do canal A
    try {
      urlA = iBusURLFactory.create(channelAName, channelAVer, channelASubject);
    }
    catch (Exception e) {
      // Coloca no log uma mensagem e termina a aplicacao (exit code=1)
      log.panic("main: iBusURLFactory.create failed for channel A: " + e);
    }


    /* ***********************************************************************
     * Define a aplicacao como um consumidor do canal A.
     *
     */

    // Instancia o objecto que recebe as mensagens enviadas em A
    ObjectReceiver rcvA = new ObjectReceiver(log, address);

    // Conecta-se ao canal A como assinante (provoca uma mudanca de vista em A)
    try {
      stackA.subscribe(urlA, rcvA);
    }
    catch (Exception e) {
      log.panic("main: stack.subscribe failed for channel A: " + e);
    }


    /* ***********************************************************************
     * Inicio da aplicacao
     *
     */

    System.out.println(Application.getAppName() + ": inicio da aplicacao.\n");

    stackV.waitTillExit();			// Aguarda por mensagens em V
  }
}



/* ***************************************************************************
 * Classe: ObjectArchiver
 *
 */
class ObjectReceiver implements iBus.Receiver {
  Log log;						// Objecto log
  String address = new String();
  static int count = 0;

  // Constructor
  public ObjectReceiver(Log log_, String address_) {
    log = log;
    address = address_;
  }

  // O metodo dispatchPush e' invocado quando uma mensagem e' recebida
  public void dispatchPush(iBusURL channel, Posting posting) {
    System.err.println("ObjectReceiver.dispatchPush called");
  }

  // O metodo dispatchPull e' invocado quando uma mensagem e' enviada
  public Posting dispatchPull(iBusURL channel, Posting request) {
    System.err.println((count++) + "\tdispatchPull: got a pull request");

    // Obtem a data e hota actual
    GenerateTick gt = new GenerateTick();
    Tick tick = gt.GetTick();
    tick.setSenderId(address);
    tick.setOrderNumber(count);

    // Instancia o objecto de resposta
    final Posting reply = new Posting();
    reply.setLength(1);
    reply.setObject(0, tick);

    // Devolve a resposta ao cliente
    return reply;
  }

  // O metodo error e' invocado quando ocorre um erro
  public void error(iBusURL channel, String details) {
    log.panic("ObjectReceiver: internal error: " + details);
  }
}



/* ***************************************************************************
 * Classe: GenerateTick
 *
 * A classe GenerateTick funciona como um gerador aleatorio de Ticks.
 * Os Ticks devolvidos podem ser do tipo News ou do tipo Financial.
 */
class GenerateTick {
  public Tick GetTick() {
    // Simulacao para objecto simples
    //return(new Ticker.News());

    // Simulacao para objecto composto
    return(new Ticker.Financial());
  }
}

// fim: Consumer.java
